#include <QString>
#include <QFile>
#include <QFileInfo>
#include <QCoreApplication>
#include <QSysInfo>

#if defined(Q_WS_X11) || defined(Q_WS_MAC)
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#endif

#ifdef Q_WS_WIN
#include <windows.h>
#endif

#include "systeminfo.h"
#include "psioptions.h"

SystemInfo::SystemInfo() : QObject(QCoreApplication::instance())
{
	// Initialize
	timezone_offset_ = 0;
	timezone_str_ = "N/A";
	os_str_ = "Unknown";
	
	// Detect
#if defined(Q_WS_X11) || defined(Q_WS_MAC)
	time_t x;
	time(&x);
	char str[256];
	char fmt[32];
	strcpy(fmt, "%z");
	strftime(str, 256, fmt, localtime(&x));
	if(strcmp(fmt, str)) {
		QString s = str;
		if(s.at(0) == '+')
			s.remove(0,1);
		s.truncate(s.length()-2);
		timezone_offset_ = s.toInt() * 60;	// FIX-ME: should really read the offset in minutes
	}
	strcpy(fmt, "%Z");
	strftime(str, 256, fmt, localtime(&x));
	if(strcmp(fmt, str))
		timezone_str_ = str;
#endif
#if defined(Q_WS_X11)
	struct utsname u;
	uname(&u);
	os_str_.sprintf("%s", u.sysname);

	// get description about os
	enum LinuxName {
		LinuxNone = 0,

		LinuxMandrake,
		LinuxDebian,
		LinuxRedHat,
		LinuxGentoo,
		LinuxSlackware,
		LinuxSuSE,
		LinuxConectiva,
		LinuxCaldera,
		LinuxLFS,

		LinuxASP, // Russian Linux distros
		LinuxALT,

		LinuxPLD, // Polish Linux distros
		LinuxAurox,
		LinuxArch
	};

	enum OsFlags {
		OsUseName = 0,
		OsUseFile,
		OsAppendFile
	};

	struct OsInfo {
		LinuxName id;
		OsFlags flags;
		QString file;
		QString name;
	} osInfo[] = {
		{ LinuxMandrake,	OsUseFile,	"/etc/mandrake-release",	"Mandrake Linux"	},
		{ LinuxDebian,		OsAppendFile,	"/etc/debian_version",		"Debian GNU/Linux"	},
		{ LinuxGentoo,		OsUseFile,	"/etc/gentoo-release",		"Gentoo Linux"		},
		{ LinuxSlackware,	OsAppendFile,	"/etc/slackware-version",	"Slackware Linux"	},
		{ LinuxPLD,		OsUseFile,	"/etc/pld-release",		"PLD Linux"		},
		{ LinuxAurox,		OsUseName,	"/etc/aurox-release",		"Aurox Linux"		},
		{ LinuxArch,		OsUseFile,	"/etc/arch-release",		"Arch Linux"		},
		{ LinuxLFS,		OsAppendFile,	"/etc/lfs-release",		"LFS Linux"		},

		// untested
		{ LinuxSuSE,		OsUseFile,	"/etc/SuSE-release",		"SuSE Linux"		},
		{ LinuxConectiva,	OsUseFile,	"/etc/conectiva-release",	"Conectiva Linux"	},
		{ LinuxCaldera,		OsUseFile,	"/etc/.installed",		"Caldera Linux"		},

		// many distros use the /etc/redhat-release for compatibility, so RedHat will be the last :)
		{ LinuxRedHat,		OsUseFile,	"/etc/redhat-release",		"RedHat Linux"		},

		{ LinuxNone,		OsUseName,	"",				""			}
	};

	for (int i = 0; osInfo[i].id != LinuxNone; i++) {
		QFileInfo fi( osInfo[i].file );
		if ( fi.exists() ) {
			char buffer[128];

			QFile f( osInfo[i].file );
			f.open( QIODevice::ReadOnly );
			f.readLine( buffer, 128 );
			QString desc(buffer);

			desc = desc.stripWhiteSpace ();

			switch ( osInfo[i].flags ) {
				case OsUseFile:
					os_str_ = desc;
					break;
				case OsUseName:
					os_str_ = osInfo[i].name;
					break;
				case OsAppendFile:
					os_str_ = osInfo[i].name + " (" + desc + ")";
					break;
			}

			break;
		}
	}
#elif defined(Q_WS_MAC)
	QSysInfo::MacVersion v = QSysInfo::MacintoshVersion;
	if(v == QSysInfo::MV_10_3)
		os_str_ = "Mac OS X 10.3";
	else if(v == QSysInfo::MV_10_4)
		os_str_ = "Mac OS X 10.4";
	else if(v == QSysInfo::MV_10_5)
		os_str_ = "Mac OS X 10.5";
	else
		os_str_ = "Mac OS X";
#endif

#if defined(Q_WS_WIN)
	TIME_ZONE_INFORMATION i;
	//GetTimeZoneInformation(&i);
	//timezone_offset_ = (-i.Bias) / 60;
	memset(&i, 0, sizeof(i));
	bool inDST = (GetTimeZoneInformation(&i) == TIME_ZONE_ID_DAYLIGHT);
	int bias = i.Bias;
	if(inDST)
		bias += i.DaylightBias;
	timezone_offset_ = -bias;
	timezone_str_ = "";
	for(int n = 0; n < 32; ++n) {
		int w = inDST ? i.DaylightName[n] : i.StandardName[n];
		if(w == 0)
			break;
		timezone_str_ += QChar(w);
	}

	QSysInfo::WinVersion v = QSysInfo::WindowsVersion;
	if(v == QSysInfo::WV_95)
		os_str_ = "Windows 95";
	else if(v == QSysInfo::WV_98)
		os_str_ = "Windows 98";
	else if(v == QSysInfo::WV_Me)
		os_str_ = "Windows Me";
	else if(v == QSysInfo::WV_DOS_based)
		os_str_ = "Windows 9x/Me";
	else if(v == QSysInfo::WV_NT)
		os_str_ = "Windows NT 4.x";
	else if(v == QSysInfo::WV_2000)
		os_str_ = "Windows 2000";
	else if(v == QSysInfo::WV_XP)
		os_str_ = "Windows XP";
	else if(v == QSysInfo::WV_2003)
		os_str_ = "Windows Server 2003";
	else if(v == QSysInfo::WV_VISTA)
		os_str_ = "Windows Vista";
	else if(v == QSysInfo::WV_NT_based)
		os_str_ = "Windows NT";

	bool displayEdition = PsiOptions::instance()->getOption("options.osversion.display-edition").toBool();
	bool displayServicePack = PsiOptions::instance()->getOption("options.osversion.display-servicepack").toBool();
	os_str_ = getWindowsVersion (displayEdition, displayServicePack);
#endif
}

#if defined(Q_WS_WIN) // Windows specific code

/*
 * michalj: Correct Windows version detection
 */

// Flags used in GetSystemMetrics (int) call
#ifndef SM_TABLETPC    // Windows XP Tablet Edition
#define SM_TABLETPC    86
#endif // #ifndef SM_TABLETPC
#ifndef SM_MEDIACENTER // Windows XP Media Center Edition
#define SM_MEDIACENTER 87
#endif // #ifndef SM_MEDIACENTER
#ifndef SM_STARTER     // Windows XP Starter Edition
#define SM_STARTER     88
#endif // #ifndef SM_STARTER
#ifndef SM_SERVERR2    // Windows Server 2003 R2
#define SM_SERVERR2    89
#endif // #ifndef SM_SERVERR2

// Flags used in OSVERSIONINFOEX.wSuiteMask field
#ifndef VER_SUITE_TERMINAL       // Terminal Services is installed. This value is always set.
#define VER_SUITE_TERMINAL       0x00000010
#endif // #ifndef VER_SUITE_TERMINAL
#ifndef VER_SUITE_EMBEDDEDNT     // Windows XP Embedded
#define VER_SUITE_EMBEDDEDNT     0x00000040
#endif // #ifndef VER_SUITE_EMBEDDEDNT
#ifndef VER_SUITE_SINGLEUSERTS   // Remote Desktop is supported, but only one interactive session is supported. This value is set unless the system is running in application server mode.
#define VER_SUITE_SINGLEUSERTS   0x00000100
#endif // #ifndef VER_SUITE_SINGLEUSERTS
#ifndef VER_SUITE_PERSONAL       // Windows XP Home Edition, Windows Vista Home Premium or Windows Vista Home Basic
#define VER_SUITE_PERSONAL       0x00000200
#endif // #ifndef VER_SUITE_PERSONAL
#ifndef VER_SUITE_BLADE          // Windows Server 2003 Web Edition
#define VER_SUITE_BLADE          0x00000400
#endif // #ifndef VER_SUITE_BLADE
#ifndef VER_SUITE_STORAGE_SERVER // Windows Storage Server 2003 or Windows Storage Server 2003 R2
#define VER_SUITE_STORAGE_SERVER 0x00002000
#endif // #ifndef VER_SUITE_STORAGE_SERVER
#ifndef VER_SUITE_COMPUTE_SERVER // Windows Server 2003, Compute Cluster Edition
#define VER_SUITE_COMPUTE_SERVER 0x00004000
#endif // #ifndef VER_SUITE_COMPUTE_SERVER

// Flags used in SYSTEM_INFO.wProcessorArchitecture field
#ifndef PROCESSOR_ARCHITECTURE_AMD64 // Intel IA-64 or AMD x86-64 architecture processor
#define PROCESSOR_ARCHITECTURE_AMD64 9
#endif // #ifndef PROCESSOR_ARCHITECTURE_AMD64

// Flags returned by GetProductInfo (DWORD, DWORD, DWORD, DWORD, LPDWORD)
#ifndef PRODUCT_BUSINESS
#define PRODUCT_BUSINESS                         0x00000006 // Business Edition
#endif // #ifndef PRODUCT_BUSINESS
#ifndef PRODUCT_BUSINESS_N
#define PRODUCT_BUSINESS_N                       0x00000010 // Business Edition N
#endif // #ifndef PRODUCT_BUSINESS_N
#ifndef PRODUCT_CLUSTER_SERVER
#define PRODUCT_CLUSTER_SERVER                   0x00000012 // Cluster Server Edition
#endif // #ifndef PRODUCT_CLUSTER_SERVER
#ifndef PRODUCT_DATACENTER_SERVER
#define PRODUCT_DATACENTER_SERVER                0x00000008 // Server Datacenter Edition (full installation)
#endif // #ifndef PRODUCT_DATACENTER_SERVER
#ifndef PRODUCT_DATACENTER_SERVER_CORE
#define PRODUCT_DATACENTER_SERVER_CORE           0x0000000C // Server Datacenter Edition (core installation)
#endif // #ifndef PRODUCT_DATACENTER_SERVER_CORE
#ifndef PRODUCT_ENTERPRISE
#define PRODUCT_ENTERPRISE                       0x00000004 // Enterprise Edition
#endif // #ifndef PRODUCT_ENTERPRISE
#ifndef PRODUCT_ENTERPRISE_N
#define PRODUCT_ENTERPRISE_N                     0x0000001B // Enterprise Edition N
#endif // #ifndef PRODUCT_ENTERPRISE_N
#ifndef PRODUCT_ENTERPRISE_SERVER
#define PRODUCT_ENTERPRISE_SERVER                0x0000000A // Server Enterprise Edition (full installation)
#endif // #ifndef PRODUCT_ENTERPRISE_SERVER
#ifndef PRODUCT_ENTERPRISE_SERVER_CORE
#define PRODUCT_ENTERPRISE_SERVER_CORE           0x0000000E // Server Enterprise Edition (core installation)
#endif // #ifndef PRODUCT_ENTERPRISE_SERVER_CORE
#ifndef PRODUCT_ENTERPRISE_SERVER_IA64
#define PRODUCT_ENTERPRISE_SERVER_IA64           0x0000000F // Server Enterprise Edition for Itanium-based Systems
#endif // #ifndef PRODUCT_ENTERPRISE_SERVER_IA64
#ifndef PRODUCT_HOME_BASIC
#define PRODUCT_HOME_BASIC                       0x00000002 // Home Basic Edition
#endif // #ifndef PRODUCT_HOME_BASIC
#ifndef PRODUCT_HOME_BASIC_N
#define PRODUCT_HOME_BASIC_N                     0x00000005 // Home Basic Edition N
#endif // #ifndef PRODUCT_HOME_BASIC_N
#ifndef PRODUCT_HOME_PREMIUM
#define PRODUCT_HOME_PREMIUM                     0x00000003 // Home Premium Edition
#endif // #ifndef PRODUCT_HOME_PREMIUM
#ifndef PRODUCT_HOME_PREMIUM_N
#define PRODUCT_HOME_PREMIUM_N                   0x0000001A // Home Premium Edition N
#endif // #ifndef PRODUCT_HOME_PREMIUM_N
#ifndef PRODUCT_HOME_SERVER
#define PRODUCT_HOME_SERVER                      0x00000013 // Home Server Edition
#endif // #ifndef PRODUCT_HOME_SERVER
#ifndef PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT
#define PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT 0x0000001E // Windows Essential Business Server Management Server
#endif // #ifndef PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT
#ifndef PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING
#define PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING  0x00000020 // Windows Essential Business Server Messaging Server
#endif // #ifndef PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING
#ifndef PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY
#define PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY   0x0000001F // Windows Essential Business Server Security Server
#endif // #ifndef PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY
#ifndef PRODUCT_SERVER_FOR_SMALLBUSINESS
#define PRODUCT_SERVER_FOR_SMALLBUSINESS         0x00000018 // Server for Small Business Edition
#endif // #ifndef PRODUCT_SERVER_FOR_SMALLBUSINESS
#ifndef PRODUCT_SMALLBUSINESS_SERVER
#define PRODUCT_SMALLBUSINESS_SERVER             0x00000009 // Small Business Server
#endif // #ifndef PRODUCT_SMALLBUSINESS_SERVER
#ifndef PRODUCT_SMALLBUSINESS_SERVER_PREMIUM
#define PRODUCT_SMALLBUSINESS_SERVER_PREMIUM     0x00000019 // Small Business Server Premium Edition
#endif // #ifndef PRODUCT_SMALLBUSINESS_SERVER_PREMIUM
#ifndef PRODUCT_STANDARD_SERVER
#define PRODUCT_STANDARD_SERVER                  0x00000007 // Server Standard Edition (full installation)
#endif // #ifndef PRODUCT_STANDARD_SERVER
#ifndef PRODUCT_STANDARD_SERVER_CORE
#define PRODUCT_STANDARD_SERVER_CORE             0x0000000D // Server Standard Edition (core installation)
#endif // #ifndef PRODUCT_STANDARD_SERVER_CORE
#ifndef PRODUCT_STARTER
#define PRODUCT_STARTER                          0x0000000B // Starter Edition
#endif // #ifndef PRODUCT_STARTER
#ifndef PRODUCT_STORAGE_ENTERPRISE_SERVER
#define PRODUCT_STORAGE_ENTERPRISE_SERVER        0x00000017 // Storage Server Enterprise Edition
#endif // #ifndef PRODUCT_STORAGE_ENTERPRISE_SERVER
#ifndef PRODUCT_STORAGE_EXPRESS_SERVER
#define PRODUCT_STORAGE_EXPRESS_SERVER           0x00000014 // Storage Server Express Edition
#endif // #ifndef PRODUCT_STORAGE_EXPRESS_SERVER
#ifndef PRODUCT_STORAGE_STANDARD_SERVER
#define PRODUCT_STORAGE_STANDARD_SERVER          0x00000015 // Storage Server Standard Edition
#endif // #ifndef PRODUCT_STORAGE_STANDARD_SERVER
#ifndef PRODUCT_STORAGE_WORKGROUP_SERVER
#define PRODUCT_STORAGE_WORKGROUP_SERVER         0x00000016 // Storage Server Workgroup Edition
#endif // #ifndef PRODUCT_STORAGE_WORKGROUP_SERVER
#ifndef PRODUCT_UNDEFINED
#define PRODUCT_UNDEFINED                        0x00000000 // An unknown product
#endif // #ifndef PRODUCT_UNDEFINED
#ifndef PRODUCT_UNLICENSED
#define PRODUCT_UNLICENSED                       0xABCDABCD // Product was not activated and grace period is over
#endif // #ifndef PRODUCT_UNLICENSED
#ifndef PRODUCT_ULTIMATE
#define PRODUCT_ULTIMATE                         0x00000001 // Ultimate Edition
#endif // #ifndef PRODUCT_ULTIMATE
#ifndef PRODUCT_ULTIMATE_N
#define PRODUCT_ULTIMATE_N                       0x0000001C // Ultimate Edition N
#endif // #ifndef PRODUCT_ULTIMATE_N
#ifndef PRODUCT_WEB_SERVER
#define PRODUCT_WEB_SERVER                       0x00000011 // Web Server Edition (full installation)
#endif // #ifndef PRODUCT_WEB_SERVER
#ifndef PRODUCT_WEB_SERVER_CORE
#define PRODUCT_WEB_SERVER_CORE                  0x0000001D // Web Server Edition (core installation)
#endif // #ifndef PRODUCT_WEB_SERVER_CORE


template <typename char_type>
const size_t SystemInfo::length (const char_type * begin) const
{
	const char_type * end = begin;
	for (; *end; ++end);
	return end - begin;
}

// Fallback to old GetVersion() WinAPI call
const QString SystemInfo::getWindowsVersionByFallback (bool edition, bool servicePackAndBuild) const
{
	DWORD version = GetVersion(); // Note that this call newer fails
	DWORD majorVersion = (DWORD) LOBYTE(LOWORD(version));
	DWORD minorVersion = (DWORD) HIBYTE(LOWORD(version));
	DWORD buildVersion = 0;

	QString result = "";

	if (version < 0x80000000)
	{
		// Windows NT/2000/XP/Server 2003/Vista/Server 2008
		buildVersion = (DWORD) HIWORD(version);
		switch (majorVersion)
		{
			case 3: // Windows NT 3.1, Windows NT 3.5 and Windows NT 3.51
			case 4: // Windows NT 4.0 and Windows NT 4.5
				result = QString("Windows NT %1.%2").arg(majorVersion).arg(minorVersion);
				break;
			case 5: // Windows 2000, Windows XP and Windows Server 2003
				switch (minorVersion)
				{
					case 0: // Windows 2000
						result = "Windows 2000";
					case 1: // Windows XP
						result = "Windows XP";
					case 2: // Windows Server 2003, Windows XP x64 and Windows Server 2003 R2
						result = "Windows Server 2003 (R2)/Windows XP x64";
						break;
					default:
						result = "[Windows 2000/XP/Server 2003]";
						break;
				}
			case 6: // Windows Vista, Windows Server 2008
				switch (minorVersion)
				{
					case 0: // Windows Vista
						result = "Windows Vista/Server 2008";
						break;
					default:
						result = "[Windows Vista/Server 2008]";
						break;
				}
				break;
			default:
				result = QString("[Windows NT] %1.%2").arg(majorVersion).arg(minorVersion);
				break;
		}

		if (servicePackAndBuild)
		{
			result += QString(" (Build %1)").arg(buildVersion);
		}
	}
	else if (majorVersion < 4)
	{
		// Windows 3.1x with Win32S
		result = "Windows 3.1x with Win32S";
	}
	else
	{
		// Windows 95/98/Me
		switch (minorVersion)
		{
			case 0: // Windows 95, Windows 95 OSR 1, Windows 95 OSR 2.0, Windows 95 OSR 2.1, Windows 95 OSR 2.5
				result = "Windows 95 (OSR1/OSR2)";
				break;
			case 3: // Windows 95 OSR2
				result = "Windows 95 OSR2";
				break;
			case 10: // Windows 98, Windows 98 Second Edition
				result = "Windows 98 (Second Edition)";
				break;
			case 90: // Windows Me
				result = "Windows Me";
				break;
			default:
				result = QString("[Windows 95/98/Me] %1.%2").arg(majorVersion).arg(minorVersion);
				break;
		}
	}

	return result;
}


typedef void (WINAPI *PGNSI) (LPSYSTEM_INFO);
typedef BOOL (WINAPI *PGPI) (DWORD, DWORD, DWORD, DWORD, PDWORD);


// Use GetProductInfo (DWORD, DWORD, DWORD, DWORD, PDWORD) WinAPI call
const QString SystemInfo::getWindowsEdition (OSVERSIONINFOEX osvi) const
{
	QString result = "";
	if ((osvi.dwMajorVersion == 6 && osvi.dwMinorVersion >= 0) || osvi.dwMajorVersion > 6)
	{
		PGPI pGPI = (PGPI) GetProcAddress (GetModuleHandle ((LPCTSTR)(const TCHAR *) "kernel32.dll"), (LPCSTR)(const char *) "GetProductInfo");
		if (pGPI)
		{
			DWORD productType;
			pGPI (osvi.dwMajorVersion, osvi.dwMinorVersion, osvi.wServicePackMajor, osvi.wServicePackMinor, &productType);
			switch (productType)
			{
				case PRODUCT_BUSINESS:
					result += " Business Edition";
					break;
				case PRODUCT_BUSINESS_N:
					result += " Business Edition N";
					break;
				case PRODUCT_CLUSTER_SERVER:
					result += " Cluster Server Edition";
					break;
				case PRODUCT_DATACENTER_SERVER:
					result += " Datacenter Server Edition (full installation)";
					break;
				case PRODUCT_DATACENTER_SERVER_CORE:
					result += " Datacenter Server Edition (core installation)";
					break;
				case PRODUCT_ENTERPRISE:
					result += " Enterprise Edition";
					break;
				case PRODUCT_ENTERPRISE_N:
					result += " Enterprise Edition N";
					break;
				case PRODUCT_ENTERPRISE_SERVER:
					result += " Enterprise Server Edition (full installation)";
					break;
				case PRODUCT_ENTERPRISE_SERVER_CORE:
					result += " Enterprise Server Edition (core installation)";
					break;
				case PRODUCT_ENTERPRISE_SERVER_IA64:
					result += " Enterprise Server Edition for Itanium-based Systems";
					break;
				case PRODUCT_HOME_BASIC:
					result += " Home Basic Edition";
					break;
				case PRODUCT_HOME_BASIC_N:
					result += " Home Basic Edition N";
					break;
				case PRODUCT_HOME_PREMIUM:
					result += " Home Premium Edition";
					break;
				case PRODUCT_HOME_PREMIUM_N:
					result += " Home Premium Edition N";
					break;
				case PRODUCT_HOME_SERVER:
					result += " Home Server";
					break;
				case PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT:
					result += " Essential Business Server Management Server";
					break;
				case PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING:
					result += " Essential Business Server Messaging Server";
					break;
				case PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY:
					result += " Essential Business Server Security Server";
					break;
				case PRODUCT_SERVER_FOR_SMALLBUSINESS:
					result += " Server for Small Business Edition";
					break;
				case PRODUCT_SMALLBUSINESS_SERVER:
					result += " Small Business Server";
					break;
				case PRODUCT_SMALLBUSINESS_SERVER_PREMIUM:
					result += " Small Business Server Premium Edition";
					break;
				case PRODUCT_STANDARD_SERVER:
					result += " Standard Server Edition (full installation)";
					break;
				case PRODUCT_STANDARD_SERVER_CORE:
					result += " Standard Server Edition (core installation)";
					break;
				case PRODUCT_STARTER:
					result += " Starter Edition";
					break;
				case PRODUCT_STORAGE_ENTERPRISE_SERVER:
					result += " Storage Server Enterprise Edition";
					break;
				case PRODUCT_STORAGE_EXPRESS_SERVER:
					result += " Storage Server Express Edition";
					break;
				case PRODUCT_STORAGE_STANDARD_SERVER:
					result += " Storage Server Standard Edition";
					break;
				case PRODUCT_STORAGE_WORKGROUP_SERVER:
					result += " Storage Server Workgroup Edition";
					break;
				case PRODUCT_UNDEFINED:
					result += " (undefined product)"; // An unknown product
					break;
				case PRODUCT_ULTIMATE:
					result += " Ultimate Edition";
					break;
				case PRODUCT_ULTIMATE_N:
					result += " Ultimate Edition N";
					break;
				case PRODUCT_WEB_SERVER:
					result += " Web Server Edition (full installation)";
					break;
				case PRODUCT_WEB_SERVER_CORE:
					result += " Web Server Edition (core installation)";
					break;
				case PRODUCT_UNLICENSED:
					result += " (unactivated, beyond grace period)";
					break;
				default:
					result += " (unknown edition)";
					break;
			}
		}
	}
	return result;
}


// Use newer GetVersionEx(LPOSVERSIONINFO) or GetVersionEx(LPOSVERSIONINFOEX) WinAPI calls
const QString SystemInfo::getWindowsVersion (bool edition, bool servicePackAndBuild) const
{
#ifdef UNICODE
#define STRLEN wcslen
#else
#define STRLEN strlen
#endif // #ifdef UNICODE
#define BUFSIZE 60

	OSVERSIONINFOEX osvi;
	SYSTEM_INFO si;
	PGNSI pGNSI;
	BOOL bOsVersionInfoEx;

	ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	if (!(bOsVersionInfoEx = GetVersionEx((OSVERSIONINFO *) &osvi)))
	{
		osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
		if (!GetVersionEx((OSVERSIONINFO *) &osvi))
			return getWindowsVersionByFallback(edition, servicePackAndBuild);
	}

	QString osvi_szCSDVersion((const QChar *) osvi.szCSDVersion, length<TCHAR>((const TCHAR *) osvi.szCSDVersion));

	QString result = "";
	switch (osvi.dwPlatformId)
	{
		case VER_PLATFORM_WIN32_NT: // Windows NT/2000/XP/Server 2003/Vista/Server 2008
			// Use GetProcAddress to avoid load issues on Windows 2000
			pGNSI = (PGNSI) GetProcAddress (GetModuleHandle ((LPCTSTR)(const TCHAR *) "kernel32.dll"), (LPCSTR)(const char *) "GetNativeSystemInfo");
			if (pGNSI)
			{
				pGNSI (&si);
			}

			// Test for the specific product
			if (osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 0)
			{
				if (osvi.wProductType == VER_NT_WORKSTATION)
				{
					result = "Windows Vista";
				}
				else
				{
					result = "Windows Server 2008";
				}
			}
			else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2)
			{
				if (GetSystemMetrics (SM_SERVERR2))
				{
					if (osvi.wSuiteMask & VER_SUITE_STORAGE_SERVER)
					{
						result = "Windows Storage Server 2003 \"R2\",";
					}
					else
					{
						result = "Windows Server 2003 \"R2\",";
					}
				}
				else if (osvi.wProductType == VER_NT_WORKSTATION && pGNSI && si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64)
				{
					result = "Windows XP Professional x64 Edition";
				}
				else
				{
					if (osvi.wSuiteMask & VER_SUITE_STORAGE_SERVER)
					{
						result = "Windows Storage Server 2003,";
					}
					else
					{
						result = "Windows Server 2003,";
					}
				}
			}
			else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1)
			{
				result = "Windows XP";
			}
			else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0)
			{
				result = "Windows 2000";
			}
			else if (osvi.dwMajorVersion <= 4)
			{
				result = "Windows NT";
			}
			else
			{
				result = "[Windows NT]";
			}

			if (edition)
			{
				if (bOsVersionInfoEx) // Test for specific product on Windows NT 4.0 SP6 and later.
				{
					QString tmp = getWindowsEdition (osvi);
					if (tmp != "")
					{
						result += tmp;
					}
					else if (osvi.wProductType == VER_NT_WORKSTATION && (!pGNSI || si.wProcessorArchitecture != PROCESSOR_ARCHITECTURE_AMD64))
					{
						// Workstation editions - note that Windows Server 2003 and Windows Server 2008 do not belong here
						if (osvi.dwMajorVersion == 4) // Windows NT 4.0 Workstation
						{
							result += " Workstation 4.0";
						}
						else if (osvi.wSuiteMask & VER_SUITE_PERSONAL) // Windows XP Home Edition, Windows Vista Home Premium or Windows Vista Home Basic
						{
							if (osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 0) // Windows Vista Home Premium, Windows Vista Home Basic
							{
								result += " Home Premium/Basic";
							}
							else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1) // Windows XP Home Edition
							{
								result += " Home Edition";
							}
							else
							{
								result += " [Home Edition]";
							}
						}
						else if (osvi.wSuiteMask & VER_SUITE_EMBEDDEDNT) // Windows XP Embedded Edition
						{
							result += " Embedded Edition";
						}
						else if (GetSystemMetrics (SM_MEDIACENTER)) // Windows XP Media Center Edition
						{
							result += " Media Center Edition";
							// TODO: Add advanced Windows XP MCE recognition
						}
						else if (GetSystemMetrics (SM_STARTER)) // Windows XP Starter Edition
						{
							result += " Starter Edition";
						}
						else if (GetSystemMetrics (SM_TABLETPC)) // Windows XP Tablet PC Edition
						{
							result += " Tablet PC Edition";
						}
						else // Windows 2000 Professional, Windows XP Professional
						{
							result += " Professional";
						}
					}
					else if (osvi.wProductType == VER_NT_SERVER || osvi.wProductType == VER_NT_DOMAIN_CONTROLLER)
					{
						// Server editions - note that Windows XP and Windows Vista do not belong here
						if (osvi.dwMajorVersion == 6 && osvi.dwMinorVersion == 0) // Windows Server 2008
						{
							if (osvi.wSuiteMask & VER_SUITE_DATACENTER)
							{
								result += " Datacenter";
							}
							else if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
							{
								result += " Enterprise";
							}
						}
						else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2) // Windows Server 2003
						{
							if (si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_IA64) // Itanium-based editions
							{
								if (osvi.wSuiteMask & VER_SUITE_DATACENTER)
								{
									result += " Datacenter Edition for Itanium-based Systems";
								}
								else if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
								{
									result += " Enterprise Edition for Itanium-based Systems";
								}
							}
							else if (si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_AMD64) // x64 editions
							{
								if (osvi.wSuiteMask & VER_SUITE_DATACENTER)
								{
									result += " Datacenter x64 Edition";
								}
								else if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
								{
									result += " Enterprise x64 Edition";
								}
								else
								{
									result += " Standard x64 Edition";
								}
							}
							else // 32-bit editions
							{
								if (osvi.wSuiteMask & VER_SUITE_DATACENTER)
								{
									result += " Datacenter Edition";
								}
								else if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
								{
									result += " Enterprise Edition";
								}
								else if (osvi.wSuiteMask & VER_SUITE_COMPUTE_SERVER)
								{
									result += " Compute Cluster Edition";
								}
								else if (osvi.wSuiteMask == VER_SUITE_BLADE)
								{
									result += " Web Edition";
								}
								else
								{
									result += " Standard Edition";
								}
							}
						}
						else if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0) // Windows 2000 Server
						{
							if (osvi.wSuiteMask & VER_SUITE_DATACENTER)
							{
								result += " Datacenter Server";
							}
							else if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
							{
								result += " Advanced Server";
							}
							else
							{
								result += " Server";
							}
						}
						else // Windows NT 4.0 Server
						{
							result += " Server 4.0";
							if (osvi.wSuiteMask & VER_SUITE_ENTERPRISE)
							{
								result += ", Enterprise Edition";
							}
						}
					}
				}
				else // Test for specific product on Windows NT 4.0 SP5 and earlier
				{
					HKEY hKey;
					char szProductType [BUFSIZE];
					DWORD dwBufLen = BUFSIZE;
					LONG lRet;

					lRet = RegOpenKeyEx (HKEY_LOCAL_MACHINE, (LPCTSTR)(const TCHAR *) "SYSTEM\\CurrentControlSet\\Control\\ProductOptions", 0, KEY_QUERY_VALUE, &hKey);
					if (lRet != ERROR_SUCCESS)
					{
						return result;
					}

					lRet = RegQueryValueEx (hKey, (LPCTSTR)(const TCHAR *) "ProductType", NULL, NULL, (LPBYTE) szProductType, &dwBufLen);
					RegCloseKey (hKey);

					if ((lRet != ERROR_SUCCESS) || (dwBufLen > BUFSIZE))
					{
						return result;
					}

					if (lstrcmpi ((LPCTSTR)(const TCHAR *) "WINNT", (LPCTSTR)(const TCHAR *) szProductType) == 0) // Windows NT 3.1 Workstation, Windows NT 3.5 Workstation, Windows NT 3.51 Workstation, Windows NT 4.0 Workstation
					{
						result += " Workstation";
					}
					else if (lstrcmpi ((LPCTSTR)(const TCHAR *) "LANMANNT", (LPCTSTR)(const TCHAR *) szProductType) == 0) // Windows NT 3.1 Server, Windows NT 3.5 Server, Windows NT 3.51 Server, Windows NT 4.0 Server
					{
						result += " Server";
					}
					else if (lstrcmpi ((LPCTSTR)(const TCHAR *) "SERVERNT", (LPCTSTR)(const TCHAR *) szProductType) == 0) // Windows NT 4.0 Advanced Server
					{
						result += " Advanced Server";
					}
					result += QString(" %1.%2").arg(osvi.dwMajorVersion).arg(osvi.dwMinorVersion);
				}

				if (servicePackAndBuild)
				{
					QString servicePackVersion = QString ("%1").arg (osvi.wServicePackMajor);
					if (osvi.wServicePackMinor)
					{
						servicePackVersion += QString (".%1").arg (osvi.wServicePackMinor);
					}
					// Display service pack (if any) and build number
					if (osvi.dwMajorVersion == 4 && (osvi_szCSDVersion == "Service Pack 6" || osvi.wServicePackMajor == 6)) // Windows NT 4.0 Service Pack 6 or Windows NT 4.0 Service Pack 6a
					{
						HKEY hKey;
						LONG lRet;

						// Test for SP6 versus SP6a.
						lRet = RegOpenKeyEx (HKEY_LOCAL_MACHINE, (LPCTSTR)(const TCHAR *) "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Hotfix\\Q246009", 0, KEY_QUERY_VALUE, &hKey);
						if (lRet == ERROR_SUCCESS) // Windows NT 4.0 Service Pack 6a
						{
							servicePackVersion = "6a";
						}
						RegCloseKey (hKey);
					}
					result += QString(" Service Pack %1 (Build %2)").arg(servicePackVersion).arg(osvi.dwBuildNumber & 0xFFFF);
				}
			}
			break;
		case VER_PLATFORM_WIN32_WINDOWS: // Windows 95/98/Me
			if (osvi.dwMajorVersion == 4 && (osvi.dwMinorVersion == 0 || osvi.dwMinorVersion == 3)) // Windows 95, Windows 95 OSR 1, Windows 95 OSR 2.0, Windows 95 OSR 2.1, Windows 95 OSR 2.5
			{
				result = "Windows 95";
				switch (LOWORD(osvi.dwBuildNumber))
				{
					case 1111: // Windows 95 OSR 2.0
						result += " OSR 2.0";
						break;
					case 1212: // Windows 95 OSR 2.1
					case 1213:
						result += " OSR 2.1";
						break;
					case 1214: // Windows 95 OSR 2.1, Windows 95 OSR 2.5
						result += " OSR 2.";
						if (osvi_szCSDVersion[1].toUpper() == 'C')
							result += "5";
						else
							result += "1";
						break;
					default:
						if (LOWORD(osvi.dwBuildNumber) > 1080) // Windows 95 OSR 2.x
						{
							result += " OSR 2.x";
						}
						else if (osvi_szCSDVersion[1].toUpper() == 'A' || osvi_szCSDVersion[1].toUpper() == 'W') // Windows 95 OSR 1
						{
							result += " OSR 1";
						}
						break;
				}
			}
			else if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 10) // Windows 98, Windows 98 Second Edition
			{
				result = "Windows 98";
				if (osvi.dwBuildNumber == 2222 || osvi_szCSDVersion[1].toUpper() == 'A' || osvi_szCSDVersion[1].toUpper() == 'B') // Windows 98 Second Edition
				{
					result += " Second Edition";
				}
			}
			else if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 90) // Windows Me
			{
				result = "Windows Me";
			}
			else
			{
				result = QString("[Windows 95/98/Me] %1.%2").arg(osvi.dwMajorVersion).arg(osvi.dwMinorVersion);
			}
			break;
		case VER_PLATFORM_WIN32s: // Windows 3.1x with Win32S
			result = "Windows 3.1x with Win32S";
			break;
	}

	return result;
}

// End of michalj changes

#endif // #if defined(Q_WS_WIN)

SystemInfo* SystemInfo::instance()
{
	if (!instance_) 
		instance_ = new SystemInfo();
	return instance_;
}

SystemInfo* SystemInfo::instance_ = NULL;

/**
 * \fn int SystemInfo::timezoneOffset()
 * \brief Local timezone offset in minutes.
 */

/**
 * \fn const QString& SystemInfo::timezoneString() const
 * \brief Local timezone name.
 */
